import { AudioContext, IAudioNode, IAudioBuffer } from 'standardized-audio-context';

interface IAdapter<Metadata> {
    validateChunk: (rawData: Uint8Array, index?: number) => boolean;
    getChunkLength: (rawData: Uint8Array, metadata: Metadata | null, index?: number) => number | null;
    getChunkDuration: (rawData: Uint8Array, metadata: Metadata | null, index?: number) => number | null;
    getChunkMetadata: (rawData: Uint8Array, index?: number) => Metadata | null;
    wrapChunk: (rawData: Uint8Array) => Uint8Array;
}

interface Loader {
    cancel(): void;
    load(opts: {
        onprogress: (progress: number, length: number, total: number) => void;
        ondata: (data: Uint8Array) => void;
        onload: () => void;
        onerror: (error: Error | ProgressEvent) => void;
    }): void;
}

declare class Clip<Metadata> {
    url: string;
    loop: boolean;
    readonly adapter: IAdapter<Metadata>;
    private callbacks;
    context: AudioContext;
    buffered: number;
    length: number;
    private _loaded;
    get loaded(): boolean;
    set loaded(value: boolean);
    private _canplaythrough;
    get canplaythrough(): boolean;
    set canplaythrough(value: boolean);
    loader: Loader;
    metadata: Metadata | null;
    playing: boolean;
    ended: boolean;
    private _startTime;
    private _currentTime;
    private __chunks;
    get _chunks(): Chunk<Metadata>[];
    set _chunks(value: Chunk<Metadata>[]);
    private _contextTimeAtStart;
    private _pendingSourceStart;
    private _connected;
    private fadeTarget;
    private _gain;
    private _loadStarted;
    private _actualPlaying;
    get stuck(): boolean;
    private _currentSource;
    private _nextSource;
    private _currentGain;
    private _nextGain;
    private _audioBufferCache;
    constructor({ context, url, loop, volume, adapter, }: {
        context?: AudioContext;
        url: string;
        loop?: boolean;
        volume?: number;
        adapter: IAdapter<Metadata>;
    });
    buffer(bufferToCompletion?: boolean): Promise<void>;
    clone(): Clone<Metadata>;
    connect(destination: IAudioNode<AudioContext>, output?: number, input?: number): this;
    disconnect(destination: IAudioNode<AudioContext>, output?: number, input?: number): void;
    _disconnectAllAndReplaceAudioContext(newAudioContext: AudioContext): void;
    dispose(): void;
    off(eventName: string, cb: (data?: any) => void): void;
    on(eventName: string, cb: (data?: any) => void): {
        cancel: () => void;
    };
    once(eventName: string, cb: (data?: any) => void): {
        cancel: () => void;
    };
    play(): Promise<unknown>;
    pause(): this;
    get currentTime(): number;
    set currentTime(currentTime: number);
    get duration(): number | null;
    get paused(): boolean;
    get volume(): number;
    set volume(volume: number);
    fade(startVolume: number, endVolume: number, duration: number): void;
    private stopFade;
    private _fire;
    private trySetupAudioBufferCache;
    private audioBufferCacheHit;
    private advanceAudioBufferCache;
    private startPlay;
    private advanceAudioNodes;
    private resetAudioNodes;
    private onCurrentSourceEnd;
    private onBufferDecoded;
    private decodeChunk;
    private tryResumePlayback;
}

declare class Chunk<Metadata> {
    clip: Clip<Metadata>;
    context: AudioContext;
    duration: number | null;
    numFrames: number | null;
    chunkIndex: number;
    raw: Uint8Array;
    extended: Uint8Array | null;
    extendedWithHeader: Uint8Array | null;
    ready: boolean;
    next: Chunk<Metadata> | null;
    readonly adapter: IAdapter<Metadata>;
    private _attached;
    private _firstByte;
    private callbacks;
    constructor({ clip, raw, onready, onerror, adapter, chunkIndex, }: {
        clip: Clip<Metadata>;
        raw: Uint8Array;
        onready: (() => void) | null;
        onerror: (error: Error) => void;
        adapter: IAdapter<Metadata>;
        chunkIndex: number;
    });
    off(eventName: string, cb: (data?: any) => void): void;
    on(eventName: string, cb: (data?: any) => void): {
        cancel: () => void;
    };
    once(eventName: string, cb: (data?: any) => void): {
        cancel: () => void;
    };
    private _fire;
    attach(nextChunk: Chunk<Metadata> | null): void;
    createBufferCallback(callback: (buffer: IAudioBuffer) => void, errback: (error: Error) => void): void;
    createBuffer(): Promise<IAudioBuffer>;
    private _ready;
}

declare class Clone<Metadata> extends Clip<Metadata> {
    original: Clip<Metadata>;
    constructor(original: Clip<Metadata>);
    buffer(): Promise<void>;
    clone(): Clone<Metadata>;
    get canplaythrough(): boolean;
    set canplaythrough(_: boolean);
    get loaded(): boolean;
    set loaded(_: boolean);
    get _chunks(): Chunk<Metadata>[];
    set _chunks(_: Chunk<Metadata>[]);
}

interface Metadata {
    mpegVersion: number;
    mpegLayer: number;
    sampleRate: number;
    channelMode: string;
}

declare const mp3Adapter: IAdapter<Metadata>;

export { Clip, mp3Adapter };
//# sourceMappingURL=types.d.ts.map
